"""
ADM Reader
"""

from struct import unpack, unpack_from, calcsize
from io import BytesIO
from collections import namedtuple
from typing import Iterable, Tuple

from lxml import etree as ET

ChannelEntry = namedtuple('ChannelEntry', "track_index uid track_ref pack_ref")

class WavADMReader:
    """
    Reads XML data from an EBU ADM (Audio Definiton Model) WAV File.
    """

    def __init__(self, axml_data: bytes, chna_data: bytes):
        header_fmt = "<HH"
        uid_fmt = "<H12s14s11sx"

        #: An :mod:`lxml.etree` of the ADM XML document
        self.axml = ET.parse(BytesIO(axml_data))

        _, uid_count = unpack(header_fmt, chna_data[0:4])

        #: A list of :class:`ChannelEntry` objects parsed from the
        #: `chna` metadata chunk. 
        self.channel_uids = []

        offset = calcsize(header_fmt)
        for _ in range(uid_count):

            track_index, uid, track_ref, pack_ref = unpack_from(uid_fmt, chna_data, offset)

            # these values are either ascii or all null

            self.channel_uids.append(ChannelEntry(track_index,
                uid.decode('ascii') , track_ref.decode('ascii'), pack_ref.decode('ascii')))

            offset += calcsize(uid_fmt)


    def track_info(self, index):
        """
        Information about a track in the WAV file.

        :param index: index of audio track (indexed from zero) 
        :returns: a dictionary with content_name, object_name, pack_format_name, pack_type, 
                  channel_format_name
        """
        channel_info = next((x for x in self.channel_uids if x.track_index == index + 1), None)
        
        if channel_info is None:
            return None

        ret_dict = {}

        nsmap = self.axml.getroot().nsmap

        trackformat_elem = self.axml.find(".//audioFormatExtended/audioTrackFormat[@audioTrackFormatID='%s']" % channel_info.track_ref, namespaces=nsmap)

        stream_id = trackformat_elem[0].text

        channelformatref_elem = self.axml.find(".//audioFormatExtended/audioStreamFormat[@audioStreamFormatID='%s']/audioChannelFormatIDRef" % stream_id, namespaces=nsmap)
        channelformat_id = channelformatref_elem.text

        packformatref_elem = self.axml.find(".//audioFormatExtended/audioStreamFormat[@audioStreamFormatID='%s']/audioPackFormatIDRef" % stream_id, namespaces=nsmap)
        packformat_id = packformatref_elem.text

        channelformat_elem = self.axml.find(".//audioFormatExtended/audioChannelFormat[@audioChannelFormatID='%s']" % channelformat_id, namespaces=nsmap)
        ret_dict['channel_format_name'] = channelformat_elem.get("audioChannelFormatName")

        packformat_elem = self.axml.find(".//audioFormatExtended/audioPackFormat[@audioPackFormatID='%s']" % packformat_id, namespaces=nsmap)
        ret_dict['pack_type'] = packformat_elem.get("typeDefinition")
        ret_dict['pack_format_name'] = packformat_elem.get("audioPackFormatName")

        object_elem = self.axml.find(".//audioFormatExtended/audioObject[audioPackFormatIDRef = '%s']" % packformat_id, namespaces=nsmap)
        ret_dict['audio_object_name'] = object_elem.get("audioObjectName")
        object_id = object_elem.get("audioObjectID")

        content_elem = self.axml.find(".//audioFormatExtended/audioContent/[audioObjectIDRef = '%s']" % object_id, namespaces=nsmap)
        ret_dict['content_name'] = content_elem.get("audioContentName")

        return ret_dict

    def to_dict(self):
        return dict(channel_entries=list(map(lambda z: z._asdict(), self.channel_uids)))


from setuptools import setup
from wavinfo import __author__, __license__, __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='wavinfo',
      version=__version__,
      author=__author__,
      author_email='jamiehardt@me.com',
      description='Probe WAVE Files for iXML, Broadcast-WAVE and other metadata.',
      long_description_content_type="text/markdown",
      long_description=long_description,
      license=__license__,
      url='https://github.com/iluvcapra/wavinfo',
      project_urls={
          'Source':
              'https://github.com/iluvcapra/wavinfo',
          'Documentation':
              'https://wavinfo.readthedocs.io/',
          'Issues':
              'https://github.com/iluvcapra/wavinfo/issues',
      },
      packages=['wavinfo'],
      classifiers=['Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: MIT License',
          'Topic :: Multimedia',
          'Topic :: Multimedia :: Sound/Audio',
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: 3.11"],
      keywords='waveform metadata audio ebu smpte avi library film tv editing editorial',
      install_requires=['lxml'],
      entry_points={
          'console_scripts': [
              'wavinfo = wavinfo.__main__:main'
          ]
      }
      )



#-*- coding: utf-8 -*-
import struct
import os
from collections import namedtuple

import pathlib

from .riff_parser import parse_chunk, ChunkDescriptor, ListChunkDescriptor
from .wave_ixml_reader import WavIXMLFormat
from .wave_bext_reader import WavBextReader
from .wave_info_reader import WavInfoChunkReader
from .wave_adm_reader import WavADMReader

#: Calculated statistics about the audio data.
WavDataDescriptor = namedtuple('WavDataDescriptor', 'byte_count frame_count')

#: The format of the audio samples.
WavAudioFormat = namedtuple('WavAudioFormat',
                            'audio_format channel_count sample_rate byte_rate block_align bits_per_sample')


class WavInfoReader:
    """
    Parse a WAV audio file for metadata.
    """

    def __init__(self, path, info_encoding='latin_1', bext_encoding='ascii'):
        """
        Create a new reader object.

        :param path: 
            A filesystem path to the wav file you wish to probe or a 
            file handle to an open file.

        :param info_encoding: 
            The text encoding of the INFO metadata fields.
            latin_1/Win CP1252 has always been a pretty good guess for this.

        :param bext_encoding: 
            The text encoding to use when decoding the string
            fields of the Broadcast-WAV extension. Per EBU 3285 this is ASCII
            but this parameter is available to you if you encounter a weirdo.


        """
        
        self.info_encoding = info_encoding
        self.bext_encoding = bext_encoding
        
        if hasattr(path, 'read'):
            self.get_wav_info(path)
            self.url = 'about:blank'
            self.path = repr(path)
        else:
            absolute_path = os.path.abspath(path)

            #: `file://` url for the file.
            self.url = pathlib.Path(absolute_path).as_uri()

            # for __repr__()
            self.path = absolute_path
            
            with open(path, 'rb') as f:
                self.get_wav_info(f)
            
    def get_wav_info(self, wavfile):
        chunks = parse_chunk(wavfile)

        self.main_list = chunks.children
        wavfile.seek(0)

        #: :class:`wavinfo.wave_reader.WavAudioFormat`
        self.fmt = self._get_format(wavfile)

        #: :class:`wavinfo.wave_bext_reader.WavBextReader` with Broadcast-WAV metadata
        self.bext = self._get_bext(wavfile, encoding=self.bext_encoding)

        #: :class:`wavinfo.wave_ixml_reader.WavIXMLFormat` with iXML metadata
        self.ixml = self._get_ixml(wavfile)

        #: :class:`wavinfo.wave_axml_reader.WavAxmlReader` with ADM metadata
        self.adm  = self._get_adm(wavfile)

        #: :class:`wavinfo.wave_info_reader.WavInfoChunkReader` with RIFF INFO metadata
        self.info = self._get_info(wavfile, encoding=self.info_encoding)
        self.data = self._describe_data()

    def _find_chunk_data(self, ident, from_stream, default_none=False):
        top_chunks = (chunk for chunk in self.main_list if type(chunk) is ChunkDescriptor and chunk.ident == ident)
        chunk_descriptor = next(top_chunks, None) if default_none else next(top_chunks)
        return chunk_descriptor.read_data(from_stream) if chunk_descriptor else None

    def _describe_data(self):
        data_chunk = next(c for c in self.main_list if type(c) is ChunkDescriptor and c.ident == b'data')

        return WavDataDescriptor(byte_count=data_chunk.length,
                                 frame_count=int(data_chunk.length / self.fmt.block_align))

    def _get_format(self, f):
        fmt_data = self._find_chunk_data(b'fmt ', f)

        # The format chunk is
        # audio_format    U16
        # channel_count   U16
        # sample_rate     U32   Note an integer
        # byte_rate       U32   == SampleRate * NumChannels * BitsPerSample/8
        # block_align     U16   == NumChannels * BitsPerSample/8
        # bits_per_sampl  U16
        packstring = "<HHIIHH"
        rest_starts = struct.calcsize(packstring)

        unpacked = struct.unpack(packstring, fmt_data[:rest_starts])

        # 0x0001	WAVE_FORMAT_PCM	PCM
        # 0x0003	WAVE_FORMAT_IEEE_FLOAT	IEEE float
        # 0x0006	WAVE_FORMAT_ALAW	8-bit ITU-T G.711 A-law
        # 0x0007	WAVE_FORMAT_MULAW	8-bit ITU-T G.711 µ-law
        # 0xFFFE	WAVE_FORMAT_EXTENSIBLE	Determined by SubFormat

        # https://sno.phy.queensu.ca/~phil/exiftool/TagNames/RIFF.html
        return WavAudioFormat(audio_format=unpacked[0],
                              channel_count=unpacked[1],
                              sample_rate=unpacked[2],
                              byte_rate=unpacked[3],
                              block_align=unpacked[4],
                              bits_per_sample=unpacked[5]
                              )

    def _get_info(self, f, encoding):
        finder = (chunk.signature for chunk in self.main_list if type(chunk) is ListChunkDescriptor)

        if b'INFO' in finder:
            return WavInfoChunkReader(f, encoding)

    def _get_bext(self, f, encoding):
        bext_data = self._find_chunk_data(b'bext', f, default_none=True)
        return WavBextReader(bext_data, encoding) if bext_data else None

    def _get_adm(self, f):
        axml = self._find_chunk_data(b'axml', f, default_none=True)
        chna = self._find_chunk_data(b'chna', f, default_none=True)
        return WavADMReader(axml_data=axml, chna_data=chna) if axml and chna else None

    def _get_ixml(self, f):
        ixml_data = self._find_chunk_data(b'iXML', f, default_none=True)
        return None if ixml_data is None else WavIXMLFormat(ixml_data.rstrip(b'\0'))

    def walk(self):
        """
        Walk all of the available metadata fields.

        :yields: a string, the :scope: of the metadatum, the string :name: of the
        metadata field, and the value.

        
        """

        scopes = ('fmt', 'data', 'ixml', 'bext', 'info', 'adm')

        for scope in scopes:
            if scope in ['fmt', 'data']:
                attr = self.__getattribute__(scope)
                for field in attr._fields:
                    yield scope, field, attr.__getattribute__(field)

            else:
                dict = self.__getattribute__(scope).to_dict() if self.__getattribute__(scope) else {}
                for key in dict.keys():
                    yield scope, key, dict[key]
        

    def __repr__(self):
        return 'WavInfoReader({}, {}, {})'.format(self.path, self.info_encoding, self.bext_encoding)
